package com.fundi.framework.common.displaylist;

/*
 *  Common Services Library - Framework plug-in.
 *  (C) Copyright Fundi Software 2011, 2016.  Fundi Confidential.
 *  All rights reserved.  Use, duplication or disclosure restricted. 
 */

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.StringTokenizer;

import javax.xml.bind.annotation.XmlRootElement;

import com.fundi.framework.common.cslcomms.CSLField;
import com.fundi.framework.common.cslcomms.CSLFieldDefn;
import com.fundi.framework.common.cslcomms.CSLObject;
import com.fundi.framework.common.instance.SerializableAppInstance;
import com.fundi.framework.common.nl1.Messages;
import com.fundi.framework.common.util.StringEditor;

@XmlRootElement
public class DisplayList implements Serializable {

	//Eye catchers.
	public static final String EC_COPYRIGHT = "(C) Copyright 2011, 2016 Fundi Software.  Fundi Confidential."; //$NON-NLS-1$
	public static final String EC_APAR = "3.1.1.00_00014"; //$NON-NLS-1$
	public static final String EC_FMID = "H30SA10"; //$NON-NLS-1$
	public static final String EC_PID = "5655-CSL"; //$NON-NLS-1$
	
	private static final long serialVersionUID = 1L;
	
	private ArrayList<ListFieldDefn> fieldDefns = new ArrayList<ListFieldDefn>();
	private ArrayList<String> keyFieldLabels = new ArrayList<String>();
	private ArrayList<String> ignoreFieldLabels = new ArrayList<String>();
	private ArrayList<String> hideFieldLabels = new ArrayList<String>();
	private ArrayList<ListRow> rows = new ArrayList<ListRow>();
	private boolean showCommonFieldsOnly = false;  //If list is built from multiple sources, only hold fields common to each source?
	private String containValue = ""; //$NON-NLS-1$
	private boolean compareRows = false;
	private int nextRowUid = 0;
	private boolean ascending = true;
	private String defaultKeyFieldLabels = null;
	private boolean defaultKeysSaved = false;
	private ArrayList<String> defaultFieldSequence = new ArrayList<String>();
	private String keyFieldSeparator = Messages.getString("DisplayList_5"); //$NON-NLS-1$
	private String timeRun = ""; //$NON-NLS-1$
	
	private ArrayList<ListRow> visibleRows = new ArrayList<ListRow>();
	
	public static final int colorDifferenceDefaultRed = 219;
	public static final int colorDifferenceDefaultGreen = 122;
	public static final int colorDifferenceDefaultBlue = 147;
	public int colorDifferenceRed = colorDifferenceDefaultRed;
	public int colorDifferenceGreen = colorDifferenceDefaultGreen;
	public int colorDifferenceBlue = colorDifferenceDefaultBlue;
	
	public static final int colorSeparatorDefaultRed = 238;
	public static final int colorSeparatorDefaultGreen = 232;
	public static final int colorSeparatorDefaultBlue = 170;
	public int colorSeparatorRed = colorSeparatorDefaultRed;
	public int colorSeparatorGreen = colorSeparatorDefaultGreen;
	public int colorSeparatorBlue = colorSeparatorDefaultBlue;
	
	public static final int colorKeyDefaultRed = 176;
	public static final int colorKeyDefaultGreen = 196;
	public static final int colorKeyDefaultBlue = 222;
	public int colorKeyRed = colorKeyDefaultRed;
	public int colorKeyGreen = colorKeyDefaultGreen;
	public int colorKeyBlue = colorKeyDefaultBlue;
	
	public static final String nothingLabel = Messages.getString("DisplayList_1"); //$NON-NLS-1$
	
	//These are for Comparator use.
	private boolean showAllKeys = true;
	private boolean showMatchedKeys = false;
	private boolean showUnmatchedKeys = false;
	private ArrayList<Integer> comparatorSourceIds;
	
	private SerializableAppInstance instance = null;
	
	private ArrayList<ListFilter> currentFilters = new ArrayList<ListFilter>();
	private boolean highlightingOnly = true;
	
	private boolean hidingBlankColumns = false;
	private boolean hidingZeroValueColumns = false;
	private boolean hidingSameValueColumns = false;
	
	private String context = ""; //$NON-NLS-1$
	
	private String title = ""; //$NON-NLS-1$
	
	public class FindItem {
		String	keyword,
				columnLabel;
		
		public FindItem(String keyword, String columnLabel) {
			setKeyword(keyword);
			setColumnLabel(columnLabel);
		}

		public String getKeyword() {
			return keyword;
		}

		public void setKeyword(String keyword) {
			this.keyword = (keyword == null ? "" : keyword.trim()); //$NON-NLS-1$
		}

		public String getColumnLabel() {
			return columnLabel;
		}

		public void setColumnLabel(String columnLabel) {
			this.columnLabel = (columnLabel == null ? null : (columnLabel.isEmpty() ? null : columnLabel));
		}
		
		public boolean equals(Object o) {
			boolean 	result;
			FindItem	other;
			
			if (FindItem.class.isInstance(o)) {
				result = true;
				other = (FindItem) o;
				if (this.keyword == null) {
					result = result && other.keyword == null;
				} else {
					result = result && this.keyword.equals(other.keyword);
				}
				if (this.columnLabel == null) {
					result = result && other.columnLabel == null;
				} else {
					result = result && this.columnLabel.equals(other.columnLabel);
				}
			} else {
				result = false;
			}
			
			return result;
		}
		
		public String toString() {
			String result;
			
			result = this.keyword + (this.getColumnLabel() != null ? " in " + this.getColumnLabel() : "");
			
			return result;
		}
		
		public boolean validate() {
			return !keyword.isEmpty();
		}
	}
	
	public DisplayList() {
		
		super();
		
		initialise();
	}
	
	public DisplayList(SerializableAppInstance instance) {
		
		super();
		
		this.instance = instance;
		initialise();
	}
	
	private void initialise() {

		ListFieldColor colorKey;
		ListFieldColor colorDiff;
		ListFieldColor colorSep;
		
		if (instance != null) {
			colorKey = instance.getListKeyBackground();
			colorDiff = instance.getComparatorDiffBackground();
			colorSep = instance.getComparatorSepBackground();
			colorKeyRed = colorKey.getRed();
			colorKeyGreen = colorKey.getGreen();
			colorKeyBlue = colorKey.getBlue();
			colorDifferenceRed = colorDiff.getRed();
			colorDifferenceGreen = colorDiff.getGreen();
			colorDifferenceBlue = colorDiff.getBlue();
			colorSeparatorRed = colorSep.getRed();
			colorSeparatorGreen = colorSep.getGreen();
			colorSeparatorBlue = colorSep.getBlue();
		}
		
		addFieldDefn(new ListFieldDefn(nothingLabel, ListFieldDefn.typeCharacter, ListFieldDefn.alignCentre));
	}
	
	public void clear() {
		
		int idx;
		
		for (idx = 0; idx < visibleRows.size(); idx++) {
			visibleRows.get(idx).setList(null);
		}
		visibleRows.clear();
		for (idx = 0; idx < rows.size(); idx++) {
			rows.get(idx).setList(null);
		}
		rows.clear();
		
		nextRowUid = 0;
	}
	
	public void addFieldDefns(ArrayList<ListFieldDefn> defns) {
		
		int idx;
		
		for (idx = 0; idx < defns.size(); idx++) {
			addFieldDefn(defns.get(idx));
		}
	}
	
	public void addFieldDefns(ArrayList<CSLFieldDefn> cslDefns, ArrayList<String> aliases) {

		int idxCsl;
		int idxFld;
		CSLFieldDefn cslDefn;
		ListFieldDefn defn;
		boolean fnd;
		String realName;
		String aliasName;
		int idxAlias;
		int pos;
		String aliasDefn;
		boolean someFieldsAlreadyDefined;
		boolean labelFnd;
		int idxLbl;
		ListFieldDefn lblDefn;
		
		if (fieldDefns.size() > 0) {
			someFieldsAlreadyDefined = !(fieldDefns.get(0).getLabel().equals(nothingLabel));
		}
		else {
			someFieldsAlreadyDefined = false;
		}
		
		for (idxCsl = 0; idxCsl < cslDefns.size(); idxCsl++) {
			cslDefn = cslDefns.get(idxCsl);
			fnd = false;
			for (idxFld = 0; idxFld < fieldDefns.size(); idxFld++) {
				if (fieldDefns.get(idxFld).getKeyLabel().equalsIgnoreCase(cslDefn.getKeyLabel())) {
					fnd = true;
					break;
				}
			}
			
			if (!fnd) {
				defn = new ListFieldDefn(cslDefn);
				if (aliases != null) {
					//Aliases are in the form realFieldName=aliasName
					for (idxAlias = 0; idxAlias < aliases.size(); idxAlias++) {
						aliasDefn = aliases.get(idxAlias);
						pos = aliasDefn.indexOf(Messages.getString("DisplayList_7")); //$NON-NLS-1$
						realName = aliasDefn.substring(0, pos);
						aliasName = aliasDefn.substring(++pos);
						if (defn.getKeyLabel().equalsIgnoreCase(realName)) {	//TODO possibly should be comparing label with realName
							defn.setAlias(aliasName);
							break;
						}
					}
				}
				
				if (showCommonFieldsOnly && someFieldsAlreadyDefined) {
					labelFnd = false;
					for (idxLbl = 0; idxLbl < fieldDefns.size(); idxLbl++) {
						lblDefn = fieldDefns.get(idxLbl);
						if (lblDefn.getLabel().equalsIgnoreCase(defn.getLabel()) && lblDefn.isInAllRows()) {
							labelFnd = true;
							break;
						}
					}
					//Unless its label is already known and in all rows, this field can't have been in all rows, since those added with prior fields lists didn't have it.
					defn.setInAllRows(labelFnd);  
				}
				addFieldDefn(defn);
			}
		}
		
		saveDefaultFieldSequence();
	}
	
	private void saveDefaultFieldSequence() {
		
		int idx;
		
		defaultFieldSequence.clear();
		
		for (idx = 0; idx < fieldDefns.size(); idx++) {
			defaultFieldSequence.add(fieldDefns.get(idx).getLabel());
		}
	}
	
	public void setFieldSequence(ArrayList<String> seqFields) {
		
		ArrayList<ListFieldDefn> saveDefns = new ArrayList<ListFieldDefn>();
		int idxFld;
		int idxSeq;
		ListFieldDefn defn;
		String seqLbl;
		
		for (idxFld = 0; idxFld < fieldDefns.size(); idxFld++) {
			saveDefns.add(fieldDefns.get(idxFld));
		}
		
		//Note that there may be more than one definition with the same label.
		fieldDefns.clear();
		for (idxSeq = 0; idxSeq < seqFields.size(); idxSeq++) {
			seqLbl = seqFields.get(idxSeq);
			if (seqLbl.length() > 0) {
				for (idxFld = 0; idxFld < saveDefns.size(); idxFld++) {
					defn = saveDefns.get(idxFld);
					if (defn.getLabel().equals(seqLbl)) {
						fieldDefns.add(defn);
						saveDefns.remove(idxFld);
						idxFld = -1;
					}
				}
			}
		}
		//Add any fields not already moved.
		for (idxFld = 0; idxFld < saveDefns.size(); idxFld++) {
			fieldDefns.add(saveDefns.get(idxFld));
		}
	}
	
	public void setKeyFields(String keyLabels) {
		
		setKeyFields(keyLabels, false);
	}
		
	public void setKeyFields(String keyLabels, boolean resettingLayout) {
			
		//The names must be separated by a semicolon, and in decreasing order of significance.
			
		StringTokenizer parser;
		String label;
		int idxLbl;
		int idxDefn;
		boolean fnd;
		ListFieldDefn defn;
		
		if (keyLabels == null) {
			parser = new StringTokenizer("", keyFieldSeparator); //$NON-NLS-1$
		}
		else {
			parser = new StringTokenizer(keyLabels, keyFieldSeparator);
		}
		
		//Save the default keys.
		if (!defaultKeysSaved) {
			defaultKeyFieldLabels = getKeyFieldsAsString();
			defaultKeysSaved = true;
		}
		if (getKeyFieldsAsString().equals(keyLabels)) {
			if (!resettingLayout) {
				ascending = !ascending;
			}
		}
		else {
			ascending = true;
		}
		
		keyFieldLabels.clear();
		for (idxDefn = 0; idxDefn < fieldDefns.size(); idxDefn++) {
			defn = fieldDefns.get(idxDefn);
			defn.setKey(false);
		}
		
		while (parser.hasMoreTokens()) {
			label = parser.nextToken().trim();
			fnd = false;
			for (idxLbl = 0; idxLbl < keyFieldLabels.size(); idxLbl++) {
				if (keyFieldLabels.get(idxLbl).equalsIgnoreCase(label)) {
					fnd = true;
					break;
				}
			}
			
			if (!fnd) {
				keyFieldLabels.add(label);
				//Set the key attribute on matching field definitions.  Note there may be more than one field with the same label (e.g. for aliases). 
				for (idxDefn = 0; idxDefn < fieldDefns.size(); idxDefn++) {
					defn = fieldDefns.get(idxDefn);
					if (defn.getCompareLabel().equalsIgnoreCase(label)) {
						defn.setKey(true);
					}
				}
			}
		}
	}
	
	public boolean isKeyField(String label) {
		
		return keyFieldLabels.contains(label);
	}
	
	public ArrayList<String> getKeyFields() {
		
		return keyFieldLabels;
	}
	
	public ArrayList<String> getAllFields() {
		
		ArrayList<String> labels = new ArrayList<String>();
		int idx;
		
		for (idx = 0; idx < fieldDefns.size(); idx++) {
			if (!labels.contains(fieldDefns.get(idx).getLabel())) {	//MA: added to handle fields being aliased to the same name
				labels.add(fieldDefns.get(idx).getLabel());
			}
		}
		
		return labels;
	}
	
	public ArrayList<String> getAllVisibleFields() {
		return getAllVisibleFields(false);
	}
	
	public ArrayList<String> getAllVisibleFields(boolean strict) {
		
		ArrayList<String> labels = new ArrayList<String>();
		int idx;
		
		for (idx = 0; idx < fieldDefns.size(); idx++) {
			if ((strict && !fieldDefns.get(idx).isHiddenByServer()) || (!strict && fieldDefns.get(idx).isVisible())) {
				if (!labels.contains(fieldDefns.get(idx).getLabel())) {	//MA: added to handle fields being aliased to the same name
					labels.add(fieldDefns.get(idx).getLabel());
				}
			}
		}
		
		return labels;
	}
	
	public ListFieldDefn getFieldDefn(String label) {
		
		int idx;
		ListFieldDefn defn = null;
		
		for (idx = 0; idx < fieldDefns.size(); idx++) {
			defn = fieldDefns.get(idx);
			if (fieldDefns.get(idx).getLabel().equals(label)) {
				return defn;
			}
		}
		
		return null;
	}
	
	public String getAllFieldsAsString() {
		
		StringBuffer flds = new StringBuffer(""); //$NON-NLS-1$
		int idxFld;
		
		for (idxFld = 0; idxFld < fieldDefns.size(); idxFld++) {
			if (idxFld > 0) {
				flds.append(keyFieldSeparator);
			}
			flds.append(fieldDefns.get(idxFld).getLabel());
		}
		
		return flds.toString();
	}
	
	public void setIgnoreFields(ArrayList<String> ignoreLabels) {
		
		ignoreFieldLabels = ignoreLabels;
	}
	
	public void addIgnoreField(String label) {
		
		int idx;
		
		for (idx = 0; idx < fieldDefns.size(); idx++) {
			if (fieldDefns.get(idx).getLabel().equals(label)) {
				fieldDefns.get(idx).setCompareIgnore(true);
				return;
			}
		}
	}
	
	public void setHideFields(ArrayList<String> hideLabels) {
		
		hideFieldLabels = hideLabels;
	}
	
	public String getKeyFieldsAsString() {
		
		StringBuffer flds = new StringBuffer(""); //$NON-NLS-1$
		int idxFld;
		
		for (idxFld = 0; idxFld < keyFieldLabels.size(); idxFld++) {
			if (idxFld > 0) {
				flds.append(keyFieldSeparator);
			}
			flds.append(keyFieldLabels.get(idxFld));
		}
		
		return flds.toString();
	}
	
	public void saveDefaultLayout() {
		
		saveDefaultFieldSequence();
		defaultKeysSaved = false;
		defaultKeyFieldLabels = getKeyFieldsAsString();
		defaultKeysSaved = true;
	}
	
	public void restoreDefaultLayout() {
		
		showAllColumns();
		
		setFieldSequence(defaultFieldSequence);
		
		setKeyFields(defaultKeyFieldLabels, true);
	}
		
	public void showSomeColumns(boolean showBlank, boolean showZero, boolean showSame) {
		
		int idxFld;
		ArrayList<ListFieldDefn> fields = getFieldDefns();
		ListFieldDefn defn;
		
		//Don't use defn.setVisible() here, because setVisible() determines if the column is permanently hidden, whereas the setHidden...
		//methods determine if it was temporarily hidden, and that's when it needs to be re-shown.
		for (idxFld = 0; idxFld < fields.size(); idxFld++) {
			defn = fields.get(idxFld);
			if (showBlank) {
				defn.setHiddenForBlankValue(false);
			}
			if (showZero) {
				defn.setHiddenForZeroValue(false);
			}
			if (showSame) {
				defn.setHiddenForSameValue(false);
			}
		}
		
		hidingBlankColumns = !(showBlank || !hidingBlankColumns);
		hidingZeroValueColumns = !(showZero || !hidingZeroValueColumns);
		hidingSameValueColumns = !(showSame || !hidingSameValueColumns);
	}

	public void showAllColumns() {
		
		int idxFld;
		ArrayList<ListFieldDefn> fields = getFieldDefns();
		ListFieldDefn defn;
		
		//Don't use defn.setVisible() here, because setVisible() determines if the column is permanently hidden, whereas the setHidden...
		//methods determine if it was temporarily hidden, and that's when it needs to be re-shown.
		for (idxFld = 0; idxFld < fields.size(); idxFld++) {
			defn = fields.get(idxFld);
			if (!defn.isVisible()) {
				defn.setHiddenForBlankValue(false);
				defn.setHiddenForZeroValue(false);
				defn.setHiddenForSameValue(false);
			}
		}
		
		hidingBlankColumns = false;
		hidingSameValueColumns = false;
		hidingZeroValueColumns = false;
	}

	public void addFieldDefn(ListFieldDefn defn) {
		
		/*
		 * A field can only exist once in the list rows.
		 * 
		 * The field list may be being built from multiple sources, so check the field isn't already defined for the row.
		 */
		
		int idxFld;
		boolean fnd = false;
		
		//If the field definitions only have a single field denoting no data, remove it.
		if (fieldDefns.size() > 0) {
			if (fieldDefns.get(0).getLabel().equals(nothingLabel)) {
				fieldDefns.remove(0);
			}
		}
		
		for (idxFld = 0; idxFld < fieldDefns.size(); idxFld++) {
			if (fieldDefns.get(idxFld).getKeyLabel().equalsIgnoreCase(defn.getKeyLabel())) {
				fnd = true;
				break;
			}
		}
		
		if (!fnd) {
			fieldDefns.add(defn);
			if (defn.isKey()) {
				if (keyFieldLabels.size() < defn.getKeySeq()) {
					//Don't necessarily get the key fields in sequence, so allocate the slots in the array list.
					for (idxFld = keyFieldLabels.size(); idxFld < defn.getKeySeq(); idxFld++) {
						keyFieldLabels.add(""); //$NON-NLS-1$
					}
				}
				keyFieldLabels.set(defn.getKeySeq() - 1, defn.getLabel());
			}
		}
		
		saveDefaultFieldSequence();
	}
	
	public void addRow(CSLObject obj, String dataSource) {
		
		addRow(obj, null, dataSource);
	}
	
	public void addRow(CSLObject obj, ArrayList<String> fixedFields, String dataSource) {
	
		/*
		 * It's important to only add a field if there is a definition for it.  It's possible the field list was created from multiple sources,
		 * in which case perhaps some sources had more fields than others which aren't to be displayed.
		 */

		int idxDefn;
		int idxFld;
		CSLField cslField;
		ListFieldDefn defn;
		ListRow row = new ListRow();
		String value;
		String fixedDefn;
		int pos;
		ArrayList<String> fixedNames = new ArrayList<String>();
		ArrayList<String> fixedValues = new ArrayList<String>();
		int idxFix;
		int idxRowFld;
		boolean inRow;
		 
		if (fixedFields != null) {
			//Fixed field definitions are in the form fieldName=value
			for (idxFix = 0; idxFix < fixedFields.size(); idxFix++) {
				fixedDefn = fixedFields.get(idxFix);
				pos = fixedDefn.indexOf(Messages.getString("DisplayList_9")); //$NON-NLS-1$
				fixedNames.add(fixedDefn.substring(0, pos));
				fixedValues.add(fixedDefn.substring(++pos));
				addFieldDefn(new ListFieldDefn(fixedNames.get(fixedNames.size() - 1), ListFieldDefn.typeCharacter, ListFieldDefn.alignLeft));
			}
		}
		
		for (idxFld = 0; idxFld < obj.getFields().size(); idxFld++) {
			cslField = obj.getFields().get(idxFld);
			value = cslField.getValue();
			for (idxDefn = 0; idxDefn < fieldDefns.size(); idxDefn++) {
				defn = fieldDefns.get(idxDefn);
				if (defn.getKeyLabel().equals(cslField.getKeyLabel())) {
					row.addField(new ListField(defn, value));
				}
			}
		}
		
		//Add any fixed fields.
		for (idxDefn = 0; idxDefn < fieldDefns.size(); idxDefn++) {
			defn = fieldDefns.get(idxDefn);
			for (idxFix = 0; idxFix < fixedNames.size(); idxFix++) {
				if (defn.getLabel().equals(fixedNames.get(idxFix))) {
					row.addField(new ListField(defn, fixedValues.get(idxFix)));
				}
			}
		}
		
		row.setData(obj);
		
		addRow(row, dataSource);
		
		if (showCommonFieldsOnly) {
			//Flag any fields which aren't in all rows.  Note that each row can add new fields not previously known, or invalidate prior fields
			//because they aren't in the new row.
			for (idxDefn = 0; idxDefn < fieldDefns.size(); idxDefn++) {
				defn = fieldDefns.get(idxDefn);
				inRow = false;
				for (idxRowFld = 0; idxRowFld < row.getFields().size(); idxRowFld++) {
					if (defn.getLabel().equals(row.getFields().get(idxRowFld).getLabel())) {
						inRow = true;
						break;
					}
				}
				if (!inRow) {
					defn.setInAllRows(false);
				}
			}
		}
	}
	
	public void addRow(ListRow row, String dataSource) {
		
		row.setList(this);
		if (compareRows && dataSource != null) {
			row.setDataSource(dataSource);
		}
		row.setUid(nextRowUid++);
		
		rows.add(row);
	}
	
	public ArrayList<ListRow> retrieveVisibleRows() {

		/*
		 *   NOTE:  This method only returns the visible rows from the list, in the appropriate sequence.
		 *   			 Rows may not be visible due to, for example, filters being applied.
		 */
		
		ListRow row;
		int idxFld;
		ListField field;
		boolean fnd;
		String srchVal = containValue.toLowerCase();
		int idxRow;
		
		visibleRows.clear();
		
		for (idxRow = 0; idxRow < rows.size(); idxRow++) {
			row = rows.get(idxRow);
			
			if (row.isVisible()) {
				if (containValue.length() == 0) {
					visibleRows.add(row);
				}
				else {
					fnd = false;
					for (idxFld = 0; idxFld < row.getFields().size(); idxFld++) {
						field = row.getFields().get(idxFld);
						if (field.getDefn().isInAllRows()) {
							if (field.getValue().toLowerCase().contains(srchVal)) {
								fnd = true;
							}
						}
					}
					if (fnd) {
						visibleRows.add(row);
					}
				}
			}
		}
		
		Collections.sort(visibleRows);
		
		//  Is this list providing comparator functionality?
		if (compareRows && visibleRows.size() > 0) {
			compare();
		}

		return visibleRows;
	}
	
	private void compare() {
		
		/*
		 * Mark the differences between fields on rows having the same key value.
		 * Place a blank row between the key groups.
		 */
		
		int idxFld;
		ListField field;
		int idxRow;
		String prevKeyValue = ""; //$NON-NLS-1$
		ListField prevField;
		ListFieldDefn defn;
		Integer[] keySourceCnt;
		int idxSce;
		String lastKey = ""; //$NON-NLS-1$
		ListRow row;
		int firstRow;
		int lastRow;
		int testRow;
		boolean allKeysMatched;
		String thisValue;
		String prevValue;
		
		for (idxRow = 0; idxRow < visibleRows.size(); idxRow++) {
			clearFieldColors(visibleRows.get(idxRow));
		}

		//Adjust the key groups for matched, unmatched.
		if (!showAllKeys) {
			for (idxRow = 0; idxRow < visibleRows.size(); idxRow++) {
				row = visibleRows.get(idxRow);
				if (!(row.getKeyValue().equals(lastKey))) {
					lastKey = row.getKeyValue();
					//Found a new key group, so determine whether to remove them from displaying.
					firstRow = idxRow;
					for (lastRow = (firstRow + 1); lastRow < visibleRows.size(); lastRow++) {
						if (!(visibleRows.get(lastRow).getKeyValue().equals(row.getKeyValue()))) {
							break;
						}
					}
					lastRow--;  //Decrement back to the last row in the key group;
				
					allKeysMatched = true;
					keySourceCnt = new Integer[comparatorSourceIds.size()];
					for (idxSce = 0; idxSce < comparatorSourceIds.size(); idxSce++) {
						keySourceCnt[idxSce] = 0;
					}
					for (idxSce = 0; idxSce < comparatorSourceIds.size(); idxSce++) {
						for (testRow = firstRow; testRow <= lastRow; testRow++) {
							keySourceCnt[Integer.parseInt(visibleRows.get(testRow).getDataSource()) - 1]++;
						}
					}
					for (idxSce = 1; idxSce < comparatorSourceIds.size(); idxSce++) {
						if (keySourceCnt[idxSce] != keySourceCnt[0]) {
							allKeysMatched = false;
							break;
						}
					}
						
					if ((allKeysMatched && showUnmatchedKeys) || (!allKeysMatched && showMatchedKeys)) {
						//Remove the group.
						for (testRow = lastRow; testRow >= firstRow; testRow--) {
							visibleRows.remove(testRow);
						}
						idxRow = firstRow - 1;  //The next key group to search will be at the same index in the list, since this group is removed;
					} 
				}
			}
		}
		
		if (visibleRows.size() > 0) {
			prevKeyValue = visibleRows.get(0).getKeyValue();
		}
		for (idxRow = 1; idxRow < visibleRows.size(); idxRow++) {
			if (!(visibleRows.get(idxRow).getKeyValue().equals(prevKeyValue))) {
				row = new ListRow();
				row.setList(this);
				visibleRows.add(idxRow, row);
				prevKeyValue = visibleRows.get(++idxRow).getKeyValue();  //Added a blank row, so increment position.
			}
			else {
				// If a field value is different to its predecessor, mark both as differing.
				for (idxFld = 0; idxFld < fieldDefns.size(); idxFld++) {
					defn = fieldDefns.get(idxFld);
					
					if (!(ignoreFieldLabels.contains(defn.getLabel()) || hideFieldLabels.contains(defn.getLabel()))) {
						field = visibleRows.get(idxRow).getField(defn.getLabel());
						prevField = visibleRows.get(idxRow - 1).getField(defn.getLabel());
						if (field != null) {
							if (prevField != null) {
								thisValue = field.getValue();
								prevValue = prevField.getValue();
								if (!(field.getDefn().getDataType().equals(ListFieldDefn.typeCharacter))) {  //Trim numeric or boolean fields before comparing.
									thisValue = thisValue.trim();
									prevValue = prevValue.trim();
								}
								if (!(thisValue.equals(prevValue))) {
									field.setHasDifferentValue(true);
									prevField.setHasDifferentValue(true);
								}
							}
							else {
								field.setHasDifferentValue(true);
							}
						}
						else {
							if (prevField != null) {
								prevField.setHasDifferentValue(true);
							}
						}
					}
				}
			}
		}
	}
	
	public int locate(String keyValue, boolean byRowPosition) {
		
		/*
		 *  keyValue is a comma delimited string containing the string to search for in each key field.
		 *  
		 *  The value for each key field is searched as a "starts with".
		 */
		
		int rowPos;
		int idxFld;
		StringTokenizer parser = new StringTokenizer(keyValue, Messages.getString("DisplayList_10")); //$NON-NLS-1$
		String fieldValue;
		boolean fndReqdPosn;
		ListRow row;
		
		StringBuffer sortValue = new StringBuffer(""); //$NON-NLS-1$
		ListField field;
		StringEditor edt;
		
		if (visibleRows.size() == 0) {
			return -1;
		}
		
		idxFld = -1;
		while (parser.hasMoreTokens()) {
			fieldValue = parser.nextToken().toLowerCase();
			idxFld++;
			//The field may not exist in some rows, so find the first row containing it.
			field = null;
			for (rowPos = 0; rowPos < visibleRows.size(); rowPos++) {
				row = visibleRows.get(rowPos);
				field = row.getField(keyFieldLabels.get(idxFld));
				if (field != null) {
					break;
				}
			}
			
			edt = new StringEditor(fieldValue);
			if (field.getDataType().equals(ListFieldDefn.typeCharacter)) {
				sortValue.append(edt.pad(ListRow.stringPadLength, " ", true)); //$NON-NLS-1$
			}
			else {
				if (field.getDataType().equals(ListFieldDefn.typeNumeric)) {	
					sortValue.append(edt.pad(ListRow.numericPadLength, "0", false)); //$NON-NLS-1$
				}
			}
		}
		
		fndReqdPosn = false;
		for (rowPos = 0; rowPos < visibleRows.size() && !fndReqdPosn; rowPos++) {
			row = visibleRows.get(rowPos);
			
			if (ascending) {
				if (row.getKeyValue().compareToIgnoreCase(sortValue.toString()) >= 0) {
					fndReqdPosn = true;
				}
			}
			else {
				if (row.getKeyValue().compareToIgnoreCase(sortValue.toString()) <= 0) {
					fndReqdPosn = true;
				}
			}
			
			if (fndReqdPosn) {
				break;
			}
		}
			
		
		if (rowPos >= visibleRows.size()) {
			rowPos = visibleRows.size() - 1;
		}
		else {
			if (rowPos < 0) {
				rowPos = 0;
			}
		}
		
		if (byRowPosition) {
			return rowPos;
		}
		else {
			return visibleRows.get(rowPos).getUid();
		}
	}
	
	public ListPosition find(String findValue, boolean forward, int startPosition, boolean byRowPosition) {
		return find(new FindItem(findValue, null), forward, startPosition, byRowPosition, null);
	}
	
	public ListPosition find(FindItem findItem, boolean forward, int startPosition, boolean byRowPosition, ListLayoutV1 layout) {
		int idxRow =-1;
		int startRow = -1;
		boolean fnd = false;
		int idxFld = -1;
		ArrayList<ListField> fields;
		String findValueLcase;
		ListField field = null;
		List<String>	searchLabels;
		

		findValueLcase = findItem.getKeyword().toLowerCase();
		if (!byRowPosition) {
			for (idxRow = 0; idxRow < visibleRows.size() && startRow < 0; idxRow++) {
				if (visibleRows.get(idxRow).getUid() == startPosition) {
					startRow = idxRow;
				}
			}
		}
		else {
			startRow = startPosition;
		}
		
		if (startRow < 0 || startRow > (visibleRows.size() - 1)) {
			return null;
		}
		
		if (startRow >= 0) {
			idxRow = startRow;
			searchLabels = (layout == null ? getAllVisibleFields() : layout.getAllVisibleFields(this));
			while (!fnd && idxRow < visibleRows.size() && idxRow >= 0) {
				fields = visibleRows.get(idxRow).getFields();
				for (idxFld = 0; idxFld < fields.size(); idxFld++) {
					field = fields.get(idxFld);
					if ((findItem.getColumnLabel() == null || findItem.getColumnLabel().equals(field.getLabel())) && searchLabels.contains(field.getLabel())) {
						if (field.isVisible() && field.getValue().toLowerCase().contains(findValueLcase)) {
							fnd = true;
							break;
						}
						if (findItem.getColumnLabel() != null) {
							//don't check the remaining fields, move on to the next row
							break;
						}
					}
				}
				if (!fnd) {
					if (forward) {
						idxRow++;
					}
					else {
						idxRow--;
					}
				}
			}
		}
		
		if (fnd) {
			return new ListPosition(idxRow, visibleRows.get(idxRow).getUid(), field.getLabel());
		}
		else {
			return null;
		}
	}

	public ArrayList<ListFieldDefn> getFieldDefns() {
		return getFieldDefns(true);
	}
	
	public ArrayList<ListFieldDefn> getFieldDefns(boolean mustBeInAllRows) {
		ArrayList<ListFieldDefn> retDefns = new ArrayList<ListFieldDefn>();
		int idxDefn;
		int idxKey;
		ListFieldDefn defn;
		int idxRet;
		boolean alreadyIncluded;
		String keyLabel;
		boolean keysFirst = false;
		
		if (instance != null) {
			keysFirst = instance.keepListKeysAtFront();
		}
		
		/*
		 * Return key fields first, so that they appear at the front of displayed lists.
		 * They should already be in the ArrayList in order of decreasing significance.
		 * 
		 * If only showing common fields, check the field is in each set of supplied field definitions.
		 */
		if (keysFirst) {
			for (idxKey = 0; idxKey < keyFieldLabels.size(); idxKey++) {
				keyLabel = keyFieldLabels.get(idxKey);
				for (idxDefn = 0; idxDefn < fieldDefns.size(); idxDefn++) {
					defn = fieldDefns.get(idxDefn);
					if (defn.getLabel().equalsIgnoreCase(keyLabel) && defn.isKey() && (defn.isInAllRows() || !mustBeInAllRows)) {
						//If a label is duplicated (e.g. via aliases), only return the first definition found for it.
						alreadyIncluded = false;
						for (idxRet = 0; idxRet < retDefns.size(); idxRet++) {
							if (retDefns.get(idxRet).getLabel().equals(defn.getLabel())) {
								alreadyIncluded = true;
								break;
							}
						}
						if (!alreadyIncluded) {
							retDefns.add(defn);
						}
					}
				}
			}
		}
		
		for (idxDefn = 0; idxDefn < fieldDefns.size(); idxDefn++) {
			defn = fieldDefns.get(idxDefn);
			if (defn.isInAllRows() || !mustBeInAllRows) {
			//if ((!defn.isKey()) && defn.isInAllRows()) {
				//If a label is duplicated (e.g. via aliases), only return the first definition found for it.
				alreadyIncluded = false;
				for (idxRet = 0; idxRet < retDefns.size(); idxRet++) {
					if (retDefns.get(idxRet).getLabel().equals(defn.getLabel())) {
						alreadyIncluded = true;
						break;
					}
				}
				if (!alreadyIncluded) {
					retDefns.add(defn);
				}
			}
		}
		
		return retDefns;
	}
	
	public void applyFilters(ArrayList<ListFilter> filters) {
		
		applyFilters(filters, false);
	}
	
	public void applyFilters(ArrayList<ListFilter> filters, boolean highlightOnly) {
		applyFilters(filters, highlightOnly, null);
	}
	
	public void applyFilters(ArrayList<ListFilter> filters, boolean highlightOnly, ListLayoutV1 layout) {
		/*
		 *  If highlightOnly = true, we aren't testing for reduction filters but instead are only interested
		 *  in detecting rows which can be highlighted.  Therefore in this case all rows are candidates. 
		 */
		
		int idxRow;
		int idxFld;
		int idxFlt;
		ListRow row;
		ListFilter filter;
		ListFieldColor color;
		boolean passes;
		ArrayList<String> highlightFields;
		ArrayList<String> visibleFields;
		int idxVis;
		
		visibleFields = (layout == null ? getAllVisibleFields() : layout.getAllVisibleFields(this));
		clearFilters();
		
		for (idxRow = 0; idxRow < rows.size(); idxRow++) {
			row = rows.get(idxRow);
			passes = true;
			//Get the complete field list from the display list (this object), because individual rows may not contain all fields.
			for (idxFlt = 0; idxFlt < filters.size() && passes; idxFlt++) {
				filter = filters.get(idxFlt);
				if (!highlightOnly || filter.isSummary()) {
					if (filter.isActive()) {
						passes = filter.rowPassesFilter(visibleFields, row);
					}
				}
			}

			if (passes && !highlightOnly) {
				//For highlight conditions, we stop at the first highlight obtained.
				color = new ListFieldColor();
				for (idxFlt = 0; idxFlt < filters.size() && color.getRed() < 0; idxFlt++) {
					filter = filters.get(idxFlt);
					highlightFields = filter.getHighlightFieldsInSequence();
					if (filter.isActive()) {
						for (idxFld = 0; idxFld < highlightFields.size() && color.getRed() < 0; idxFld++) {
							if (highlightFields.get(idxFld).equals(ListFilterCondition.anyField)) {
								for (idxVis = 0; idxVis < visibleFields.size() && color.getRed() < 0; idxVis++) {
									color = filter.getFieldHighlight(row.getField(visibleFields.get(idxVis)));
								}
							}
							else {
								color = filter.getFieldHighlight(row.getField(highlightFields.get(idxFld)));
							}
						}
					}
				}
				row.setHighlightColor(color);
			}

			row.setVisible(passes);
		}
		
		currentFilters.addAll(filters);
		//highlightingOnly = highlightOnly;
	}
	
	public void clearFilters() {
		
		int idxRow;
		ListRow row;
		
		for (idxRow = 0; idxRow < rows.size(); idxRow++) {
			row = rows.get(idxRow);
			row.setVisible(true);
			clearFieldColors(row);
		}
		
		currentFilters.clear();
		highlightingOnly = true;
	}
	
	private void clearFieldColors(ListRow row) {

		int idxFld;
		ArrayList<ListField> fields;

		row.resetHighlightColor();
		fields = row.getFields();
		for (idxFld = 0; idxFld < fields.size(); idxFld++) {
			fields.get(idxFld).resetColor();
		}
	}

	public void setShowCommonFieldsOnly(boolean showCommonFieldsOnly) {
		
		this.showCommonFieldsOnly = showCommonFieldsOnly;
	}

	public void setContainValue(String containValue) {
		
		this.containValue = containValue;
	}

	public boolean compareRows() {
		
		return compareRows;
	}

	public void setCompareRows(boolean compareRows) {
		
		this.compareRows = compareRows;
	}

	public boolean isAscending() {
		
		return ascending;
	}

	public void setAscending(boolean ascending) {
		
		this.ascending = ascending;
	}

	public String getTimeRun() {
		return timeRun;
	}

	public void setTimeRun(String timeRun) {
		this.timeRun = timeRun;
	}
	
	public int getVisibleRowCount() {
		
		return visibleRows.size();
	}

	public void setShowAllKeys(boolean showAllKeys) {
		
		this.showAllKeys = showAllKeys;
	}

	public void setShowMatchedKeys(boolean showMatchedKeys) {
		
		this.showMatchedKeys = showMatchedKeys;
	}

	public void setShowUnmatchedKeys(boolean showUnmatchedKeys) {
		
		this.showUnmatchedKeys = showUnmatchedKeys;
	}

	public void setComparatorSourceIds(ArrayList<Integer> comparatorSourceIds) {
		
		this.comparatorSourceIds = comparatorSourceIds;
	}

	public ArrayList<ListRow> getAllRows() {
		
		return rows;
	}

	public ArrayList<ListRow> getVisibleRows() {
		
		return visibleRows;
	}

	public SerializableAppInstance getInstance() {
		
		return instance;
	}
	
	public void setInstance(SerializableAppInstance instance) {
		
		this.instance = instance;
	}

	public String getContext() {
		 
		return context;
	}

	public void setContext(String context) {
		
		this.context = context;
	}
	
	public void removeKeyField(String label) {
		
		int idx;
		ListFieldDefn defn;
		
		keyFieldLabels.remove(label);
		for (idx = 0; idx < fieldDefns.size(); idx++) {
			defn = fieldDefns.get(idx);
			if (defn.getLabel().equals(label)) {
				defn.setKey(false);
				defn.setKey(0);
			}
		}
	}

	public ArrayList<ListFilter> getCurrentFilters() {
		
		return currentFilters;
	}
	
	public void setCurrentFilters(ArrayList<ListFilter> filters) {
		
		currentFilters.addAll(filters);
	}

	public boolean isHighlightingOnly() {
		
		return highlightingOnly;
	}

	public boolean isHidingBlankColumns() {
		return hidingBlankColumns;
	}

	public void setHidingBlankColumns(boolean hidingBlankColumns) {
		this.hidingBlankColumns = hidingBlankColumns;
	}

	public boolean isHidingZeroValueColumns() {
		return hidingZeroValueColumns;
	}

	public void setHidingZeroValueColumns(boolean hidingZeroValueColumns) {
		this.hidingZeroValueColumns = hidingZeroValueColumns;
	}

	public boolean isHidingSameValueColumns() {
		return hidingSameValueColumns;
	}

	public void setHidingSameValueColumns(boolean hidingSameValueColumns) {
		this.hidingSameValueColumns = hidingSameValueColumns;
	}

	public ArrayList<String> getDefaultKeyFieldLabels() {
		ArrayList<String> 	result;
		StringTokenizer		tokenizer;
		String				label;
		
		//Save the default keys if not saved...
		if (!defaultKeysSaved) {
			defaultKeyFieldLabels = getKeyFieldsAsString();
			defaultKeysSaved = true;
		}
		result = new ArrayList<String>();
		if (defaultKeyFieldLabels != null) {
			tokenizer = new StringTokenizer(defaultKeyFieldLabels, keyFieldSeparator);
			while (tokenizer.hasMoreTokens()) {
				label = tokenizer.nextToken();
				if (label != null) {
					result.add(label);
				}
			}
		}
		
		return result;
	}

	public ArrayList<String> getDefaultFieldSequence() {
		return defaultFieldSequence;
	}

	public String getTitle() {
		
		return title;
	}

	public void setTitle(String title) {
		
		this.title = title;
	}
}
