package com.fundi.framework.common.navigation;

/*
 *  Common Services Library - Framework plug-in.
 *  (C) Copyright Fundi Software 2011, 2016.  Fundi Confidential.
 *  All rights reserved.  Use, duplication or disclosure restricted. 
 */

import java.util.ArrayList;
import java.util.Iterator;

import com.fundi.framework.common.instance.AppInstance;
import com.fundi.framework.common.instance.Environment;

/**
 * It's worth noting here a difference between the nature of the older methods (navigationItems, userCreatableItemTypes, 
 * navigationMenuItems, contextMenuItems) and the new set (filterMenuItems, findMenuItems, listLayoutMenuItems). 
 * 
 * The first are concerned with with what a plugin might contribute to a given menu based on the resource types it 
 * understands and the functionalities it can support for a given resource. Hence, it makes sense to pass the resource
 * being examined as the sourceObject and have potentially many providers contribute items for manipulating this resource.
 * 
 * The second group are primarily concerned with the model of functionality that the UI is providing. For example filters 
 * are a framework entity unique to our plugin code which model the filtering that may be applied to resource data. Hence
 * when this group of methods is called, the need is to pass information on the UI state at a level that is common to all
 * implementations of this interface, for example, a list of search history for findMenuItems(), or a filterHelper for 
 * filterMenuItems()). In this scenario, the NavigationProvider is actually acting as a repository for implementing the 
 * common functionality of a particular UI feature in order to support that type of UI - it holds the <i>model</i> code 
 * essentially. While technically possible to allow multiple providers to contribute to these responses it is not 
 * realistically expected.
 * 
 * Part of the rationale for accumulating this function in the NavigationProvider at this time is that a performance hit
 * has historically been observed from having too many contributors to an extension point. Also, the number of methods (to
 * represent objects) is not expected to be particularly large. This is not ideal, but workable.
 */
public class NavigationProvider {

	//Eye catchers.
	public static final String EC_COPYRIGHT = "(C) Copyright 2011, 2016 Fundi Software.  Fundi Confidential."; //$NON-NLS-1$
	public static final String EC_APAR = "3.1.1.00_00014"; //$NON-NLS-1$
	public static final String EC_FMID = "H30SA10"; //$NON-NLS-1$
	public static final String EC_PID = "5655-CSL"; //$NON-NLS-1$
	
	protected AppInstance instance;
	
	private ArrayList<NavigationItem> navigationItems = new ArrayList<NavigationItem>();
	private ArrayList<NavigationType> userCreatableItemTypes = new ArrayList<NavigationType>();
	private ArrayList<MenuItem> navigationMenuItems = new ArrayList<MenuItem>();
	private ArrayList<MenuItem> contextMenuItems = new ArrayList<MenuItem>();
	private ArrayList<MenuItem> filterMenuItems = new ArrayList<MenuItem>();
	private ArrayList<MenuItem> findMenuItems = new ArrayList<MenuItem>();
	private ArrayList<MenuItem> editMenuItems = new ArrayList<MenuItem>();
	
	public NavigationProvider() {
		
		super();
	}
	
	public NavigationProvider(AppInstance instance) {
		
		super();
		
		initialise(instance);
	}
	
	public void initialise(AppInstance instance) {
		
		this.instance = instance;
	}

	public void addNavigationItem(NavigationItem newItem) {

		Iterator<NavigationItem> itItm;
		boolean fnd = false;
		NavigationItem item;
		
		for (itItm = navigationItems.iterator(); itItm.hasNext();) {
			item = itItm.next();
			if (item.equals(newItem)) {
				fnd = true;
				break;
			}
		}
		if (!fnd) {
			navigationItems.add(newItem);
		}
	}

	public ArrayList<NavigationItem> getNavigationItems() {
		
		return getNavigationItems(true);
	}
	
	public ArrayList<NavigationItem> getNavigationItems(boolean testConnections) {
		
		return navigationItems;
	}
	
	public ArrayList<NavigationType> getUserCreatableNavigationTypes() {
		
		return userCreatableItemTypes;
	}
	
	public ArrayList<MenuItem> getNavigationMenuItems(Environment environment) {
		
		return navigationMenuItems;
	}
	
	public ArrayList<MenuItem> getContextMenuItems(Environment environment, Object sourceObject, Object contextObject) {
		//need to call the old signature if this method has not been overridden in the product plug-in to maintain backwards compatibility
		return getContextMenuItems(environment, sourceObject);
	}

	public ArrayList<MenuItem> getContextMenuItems(Environment environment, Object sourceObject) {
		
		return contextMenuItems;
	}
	
	public ArrayList<MenuItem> getFilterMenuItems(Environment environment, Object sourceObject) {
		
		return filterMenuItems;
	}
	
	public ArrayList<MenuItem> getFindMenuItems(Environment environment, Object sourceObject) {
		
		return findMenuItems;
	}
	
	public ArrayList<MenuItem> getEditMenuItems(Environment environment, Object sourceObject, Object contextObject) {
		
		return editMenuItems;
	}
	
	public void run(MenuItem item, Environment environment, Object sourceObject, Object sourceViewer) {}
	
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		String className = this.getClass().getName();
		result = prime * result + ((className == null) ? 0 : className.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		return true;
	}

	
}
